package com.yonge.cooleshow.biz.dal.service.impl;

import com.alibaba.fastjson.JSONObject;
import com.baomidou.mybatisplus.core.metadata.IPage;
import com.baomidou.mybatisplus.core.toolkit.Wrappers;
import com.baomidou.mybatisplus.extension.service.impl.ServiceImpl;
import com.yonge.cooleshow.biz.dal.dto.UserAccountRecordDto;
import com.yonge.cooleshow.biz.dal.dto.req.TotalReq;
import com.yonge.cooleshow.biz.dal.entity.UserAccountRecord;
import com.yonge.cooleshow.biz.dal.entity.UserOrder;
import com.yonge.cooleshow.biz.dal.enums.InOrOutEnum;
import com.yonge.cooleshow.biz.dal.service.UserOrderService;
import com.yonge.cooleshow.biz.dal.vo.UserAccountRecordVo;
import com.yonge.cooleshow.biz.dal.vo.res.OrderPayRes;
import com.yonge.cooleshow.common.enums.CacheNameEnum;
import com.yonge.cooleshow.biz.dal.enums.FrozenTypeEnum;
import com.yonge.cooleshow.biz.dal.service.UserAccountRecordService;
import com.yonge.cooleshow.biz.dal.vo.res.AccountTotal;
import com.yonge.cooleshow.common.entity.HttpResponseResult;
import com.yonge.cooleshow.common.enums.PostStatusEnum;
import com.yonge.toolset.base.exception.BizException;
import com.yonge.toolset.payment.util.DistributedLock;
import org.redisson.api.RLock;
import org.redisson.api.RedissonClient;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;
import com.yonge.cooleshow.biz.dal.entity.UserAccount;
import com.yonge.cooleshow.biz.dal.vo.UserAccountVo;
import com.yonge.cooleshow.biz.dal.dto.search.UserAccountSearch;
import com.yonge.cooleshow.biz.dal.dao.UserAccountDao;
import com.yonge.cooleshow.biz.dal.service.UserAccountService;
import org.springframework.transaction.annotation.Transactional;

import java.math.BigDecimal;
import java.math.RoundingMode;
import java.util.Date;
import java.util.List;
import java.util.Objects;
import java.util.concurrent.*;


@Service
public class UserAccountServiceImpl extends ServiceImpl<UserAccountDao, UserAccount> implements UserAccountService {
    private final static Logger log = LoggerFactory.getLogger(UserAccountServiceImpl.class);

    @Autowired
    private UserAccountRecordService userAccountRecordService;
    @Autowired
    private RedissonClient redissonClient;
    @Autowired
    private UserOrderService orderService;

    @Override
    public UserAccountVo detail(Long id) {
        UserAccountVo detail = baseMapper.detail(id);
        if (null == detail) {
            throw new BizException("用户没有开通账户");
        }
        return detail;
    }

    @Override
    public IPage<UserAccountVo> selectPage(IPage<UserAccountVo> page, UserAccountSearch query) {
        return page.setRecords(baseMapper.selectPage(page, query));
    }

    @Override
    public HttpResponseResult<UserAccountRecord> accountRecord(UserAccountRecordDto accountRecordDto) {
        if (null == accountRecordDto.getInOrOut()
                || null == accountRecordDto.getPostStatus()
                || (!PostStatusEnum.WAIT.equals(accountRecordDto.getPostStatus()) && !PostStatusEnum.FROZEN.equals(accountRecordDto.getPostStatus()))
        ) {
            throw new BizException("记录入账-入账状态异常: param is {}", JSONObject.toJSONString(accountRecordDto));
        }

        if (null == accountRecordDto.getTransAmount() || null == accountRecordDto.getBizId()
                || null == accountRecordDto.getBizType()) {
            throw new BizException("记录入账-缺少入账参数: param is {}", JSONObject.toJSONString(accountRecordDto));
        }
        if (BigDecimal.ZERO.compareTo(accountRecordDto.getTransAmount()) < 0) {
            throw new BizException("记录入账-变动金额不能为负数: param is {}", JSONObject.toJSONString(accountRecordDto));
        }

        HttpResponseResult<UserAccountRecord> res = DistributedLock.of(redissonClient)
                .runIfLockToFunction(CacheNameEnum.LOCK_RECORD_ACCOUNT.getRedisKey(accountRecordDto.getAccountId())
                        , this::doAccountRecord, accountRecordDto, 10L);
        if (null != res) {
            return res;
        } else {
            throw new BizException("记录入账-插入账户记录失败: param is {}", JSONObject.toJSONString(accountRecordDto));
        }
    }

    @Transactional(rollbackFor = Exception.class)
    HttpResponseResult<UserAccountRecord> doAccountRecord(UserAccountRecordDto accountRecordDto) {
        //收入要校验金额
        if (InOrOutEnum.IN.equals(accountRecordDto.getInOrOut())) {
            UserOrder userOrder = orderService.getOne(Wrappers.<UserOrder>lambdaQuery()
                    .eq(UserOrder::getOrderNo, accountRecordDto.getOrderNo()));
            if (null == userOrder) {
                throw new BizException("记录入账-未找到入账对应订单信息： param is {}", JSONObject.toJSONString(accountRecordDto));
            }
            //收入判断订单金额，所有收入不能大于订单用户付款
            BigDecimal totalTransAmount = baseMapper.totalTransAmount(accountRecordDto);
            if (null == totalTransAmount) {
                totalTransAmount = BigDecimal.ZERO;
            }
            //金额校验失败
            if (null == userOrder || totalTransAmount.add(accountRecordDto.getTransAmount()).compareTo(userOrder.getActualPrice()) > 0) {
                log.error("记录入账-订单入账总金额大于购买金额： param is {}" + JSONObject.toJSONString(accountRecordDto));
                accountRecordDto.setErrFlag(1);
                accountRecordDto.setErrMsg("账户变更异常，订单入账总金额大于购买金额，订单号：" + accountRecordDto.getOrderNo());
            }
        } else {
            //支出 需要判断所有支出金额不能大于账户余额
            UserAccountVo detail = baseMapper.detail(accountRecordDto.getAccountId());
            if (detail.getAmountUsable().compareTo(accountRecordDto.getTransAmount()) < 0) {
                throw new BizException("记录入账-账户余额不足: param is {}", JSONObject.toJSONString(accountRecordDto));
            }
        }
        //如果是冻结入账，需要修改主账户冻结金额和可用余额
        if (PostStatusEnum.FROZEN.equals(accountRecordDto.getPostStatus()) && accountRecordDto.getErrFlag() == 0) {
            baseMapper.frozenChangeAccount(accountRecordDto.getAccountId(), accountRecordDto.getTransAmount(), accountRecordDto.getInOrOut().getCode());
        }
        //插入账户变更记录
        userAccountRecordService.save(accountRecordDto);
        return HttpResponseResult.succeed(accountRecordDto);
    }

    @Override
    public HttpResponseResult<UserAccountRecord> accountChange(Long recordId, PostStatusEnum postStatus) {
        if (null == postStatus || null == recordId
                || (!PostStatusEnum.RECORDED.equals(postStatus) && !PostStatusEnum.CANCEL.equals(postStatus))) {
            throw new BizException("入账状态变更, 参数异常: recordId is {} postStatus is {}", recordId, postStatus.getCode());
        }
        UserAccountRecord param = new UserAccountRecord();
        param.setId(recordId);
        param.setPostStatus(postStatus);
        HttpResponseResult<UserAccountRecord> res = DistributedLock.of(redissonClient)
                .runIfLockToFunction(CacheNameEnum.LOCK_CHANGE_ACCOUNT.getRedisKey(recordId)
                        , this::doAccountChange, param, 10L);
        if (null != res) {
            return res;
        } else {
            throw new BizException("记录变更-账户记录变更失败: param is {}", JSONObject.toJSONString(param));
        }
    }

    @Transactional(rollbackFor = Exception.class)
    public HttpResponseResult<UserAccountRecord> doAccountChange(UserAccountRecord param) {
        UserAccountRecordVo detail = userAccountRecordService.detail(param.getId());
        if (null == detail) {
            throw new BizException("入账状态变更, 未找到记录信息: recordId is {} postStatus is {}", param.getId(), param.getPostStatus().getCode());
        }
        detail.setPostStatus(param.getPostStatus());
        //待入账
        if (PostStatusEnum.WAIT.equals(detail.getPostStatus())) {
            if (PostStatusEnum.RECORDED.equals(param.getPostStatus())) {
                //正常入账
                baseMapper.changeAccount(detail.getAccountId(), detail.getTransAmount(), detail.getInOrOut().getCode());
            } else if (PostStatusEnum.CANCEL.equals(param.getPostStatus())) {
                //取消入账
            }
        } else if (PostStatusEnum.FROZEN.equals(detail.getPostStatus())) {
            if (PostStatusEnum.RECORDED.equals(param.getPostStatus())) {
                //冻结后正常入账
                baseMapper.frozenDeductChangeAccount(detail.getAccountId(), detail.getTransAmount(), detail.getInOrOut().getCode());
            } else if (PostStatusEnum.CANCEL.equals(param.getPostStatus())) {
                //冻结后取消入账
                baseMapper.frozenBackChangeAccount(detail.getAccountId(), detail.getTransAmount(), detail.getInOrOut().getCode());
            }
        }

        UserAccountVo accountVo = detail(detail.getAccountId());
        if (accountVo.getAmountTotal().doubleValue() < 0
                || accountVo.getAmountUsable().doubleValue() < 0 || accountVo.getAmountFrozen().doubleValue() < 0) {
            throw new BizException("账户变更失败");
        }
        //插入账户变更记录
        detail.setAccountBalance(accountVo.getAmountUsable());
        detail.setUpdateTime(new Date());
        userAccountRecordService.updateById(detail);
        return HttpResponseResult.succeed(detail);
    }

    @Override
    public HttpResponseResult<AccountTotal> accountTotal(TotalReq totalReq) {
        List<AccountTotal> infoList = baseMapper.accountTotal(totalReq.getTimeType().getCode(), totalReq);

        BigDecimal practiceAmount = BigDecimal.ZERO;
        BigDecimal liveAmount = BigDecimal.ZERO;
        BigDecimal videoAmount = BigDecimal.ZERO;
        BigDecimal musicAmount = BigDecimal.ZERO;
        BigDecimal liveShareAmount = BigDecimal.ZERO;
        BigDecimal videoShareAmount = BigDecimal.ZERO;
        BigDecimal musicShareAmount = BigDecimal.ZERO;
        BigDecimal vipShareAmount = BigDecimal.ZERO;
        BigDecimal mallShareAmount = BigDecimal.ZERO;

        for (AccountTotal info : infoList) {
            info.setPracticeAmount(null == info.getPracticeAmount() ? BigDecimal.ZERO : info.getPracticeAmount());
            info.setLiveAmount(null == info.getLiveAmount() ? BigDecimal.ZERO : info.getLiveAmount());
            info.setVideoAmount(null == info.getVideoAmount() ? BigDecimal.ZERO : info.getVideoAmount());
            info.setMusicAmount(null == info.getMusicAmount() ? BigDecimal.ZERO : info.getMusicAmount());

            info.setLiveShareAmount(null == info.getLiveShareAmount() ? BigDecimal.ZERO : info.getLiveShareRate());
            info.setVideoShareAmount(null == info.getVideoShareAmount() ? BigDecimal.ZERO : info.getVipShareAmount());
            info.setMusicShareAmount(null == info.getMusicShareAmount() ? BigDecimal.ZERO : info.getMusicShareAmount());
            info.setVipShareAmount(null == info.getVipShareAmount() ? BigDecimal.ZERO : info.getVipShareAmount());
            info.setMallShareAmount(null == info.getMusicShareAmount() ? BigDecimal.ZERO : info.getMusicShareAmount());

            practiceAmount = practiceAmount.add(info.getPracticeAmount());
            liveAmount = liveAmount.add(info.getLiveAmount());
            videoAmount = videoAmount.add(info.getVideoAmount());
            musicAmount = musicAmount.add(info.getMusicAmount());

            liveShareAmount = liveShareAmount.add(info.getLiveShareAmount());
            videoShareAmount = videoShareAmount.add(info.getVideoShareAmount());
            musicShareAmount = musicShareAmount.add(info.getMusicShareAmount());
            vipShareAmount = vipShareAmount.add(info.getVipShareAmount());
            mallShareAmount = mallShareAmount.add(info.getMallShareAmount());
        }
        AccountTotal total = new AccountTotal();

        total.setTotalInAmount(
                practiceAmount.add(liveAmount).add(videoAmount).add(musicAmount)
                        .add(liveShareAmount).add(videoShareAmount).add(musicShareAmount)
                        .add(vipShareAmount).add(mallShareAmount)
        );
        if (total.getTotalInAmount().doubleValue() > 0) {
            total.setPracticeAmount(practiceAmount);
            total.setPracticeRate(total.getPracticeAmount()
                    .divide(total.getTotalInAmount(), 4, RoundingMode.HALF_UP)
                    .multiply(new BigDecimal("100")));

            total.setLiveAmount(liveAmount);
            total.setLiveRate(total.getLiveAmount()
                    .divide(total.getTotalInAmount(), 4, RoundingMode.HALF_UP)
                    .multiply(new BigDecimal("100")));

            total.setVideoAmount(videoAmount);
            total.setVideoRate(total.getVideoAmount()
                    .divide(total.getTotalInAmount(), 4, RoundingMode.HALF_UP)
                    .multiply(new BigDecimal("100")));

            total.setMusicAmount(musicAmount);
            total.setMusicRate(total.getMusicAmount()
                    .divide(total.getTotalInAmount(), 4, RoundingMode.HALF_UP)
                    .multiply(new BigDecimal("100")));

            total.setLiveShareAmount(liveShareAmount);
            total.setLiveShareRate(total.getLiveShareAmount()
                    .divide(total.getTotalInAmount(), 4, RoundingMode.HALF_UP)
                    .multiply(new BigDecimal("100")));

            total.setVideoShareAmount(videoShareAmount);
            total.setVideoShareRate(total.getVideoShareAmount()
                    .divide(total.getTotalInAmount(), 4, RoundingMode.HALF_UP)
                    .multiply(new BigDecimal("100")));

            total.setMusicShareAmount(musicShareAmount);
            total.setMusicShareRate(total.getMusicShareAmount()
                    .divide(total.getTotalInAmount(), 4, RoundingMode.HALF_UP)
                    .multiply(new BigDecimal("100")));

            total.setVipShareAmount(vipShareAmount);
            total.setVipShareRate(total.getVipShareAmount()
                    .divide(total.getTotalInAmount(), 4, RoundingMode.HALF_UP)
                    .multiply(new BigDecimal("100")));

            total.setMallShareAmount(mallShareAmount);
            total.setMallShareRate(total.getMallShareAmount()
                    .divide(total.getTotalInAmount(), 4, RoundingMode.HALF_UP)
                    .multiply(new BigDecimal("100")));
        }
        total.setInfoList(infoList);
        return HttpResponseResult.succeed(total);
    }


}